extern curr		;linux
extern Round 
extern enqueue
extern dequeue
extern timer
;extern _curr		;mac
;extern _Round
;extern _enqueue
;extern _dequeue
;extern _timer

%define cur curr
%define rnd Round
%define dq  dequeue
%define nq  enqueue
%define tmr timer


;%define write_scr _write_scr
;%define pmWriteScr _pmWriteScr
;%define loadGDT _loadGDT
;%define loadIDT _loadIDT
;%define setSelectors _setSelectors
;%define go _go
;%define schedule _schedule
;%define outportb _outportb
;%define mutex_lock _mutex_lock
;%define mutex_unlock _mutex_unlock
;%define sema_up _sema_up
;%define sema_down _sema_down


segment .text
	global write_scr     ;linux
	global pmWriteScr    ;linux
	global loadGDT	     ;linux
	global loadIDT		 ;linux
	global setSelectors  ;linux
	global go
	global schedule
	global outportb
	global mutex_lock
	global mutex_unlock
	global sema_up
	global sema_down		 
	global _write_scr    ;mac
	global _pmWriteScr   ;mac
	global _loadGDT	     ;mac
	global _loadIDT		 ;mac
	global _setSelectors ;mac
	global _go
	global _schedule
	global _outportb
	global _mutex_lock
	global _mutex_unlock
	global _sema_up
	global _sema_down		 

%define	col 	[ebp+16]
%define	row 	[ebp+12]
%define	addr	[ebp+8]

write_scr:		;linux
;_write_scr:		;mac
	push ebp
	mov ebp, esp	
	push edi
	push esi
	push eax
	push ebx
	push edx
	pushf

	mov edi, 0XB8000	; the value for the video
	mov esi, addr		; the address of the string in the source

				;to satisfy ((row * 80) + col)*2, needed to add to value in edi
	mov eax, row	  	; row val
	mov ebx, col		; col val
	mov edx, 80		

	mul edx			; row*80
	add eax, ebx		; (row*80) + col in eax
	shl eax, 1		; ((row*80)+col)*2 in eax
	add edi, eax		; edi has the correct location to start
	mov al, 0X0A

_loop:
	cmp byte [esi], 0
	je exit

	movsb
	stosb
	jmp _loop
exit:

	popf
	pop edx
	pop ebx
	pop eax
	pop esi
	pop edi
	mov esp, ebp
	pop ebp
	ret

pmWriteScr:		;linux
;_pmWriteScr:		;mac
	push ebp
	mov ebp, esp
	push edi
	push esi
	push eax
	push ebx
	push edx
	pushf
	cli
	
	mov esi, addr		; the address of the string in the source

	mov eax, row
	mov ebx, col
	mov edx, 80

	mul edx
	add eax, ebx
	shl eax, 1
	mov edi, eax
	mov al, 0x0A

pmLoop:
	cmp byte [esi], 0
	je pmExit

	movsb
	stosb
	jmp pmLoop
pmExit:

	popf
	pop edx
	pop ebx
	pop eax
	pop esi
	pop edi
	mov esp, ebp
	pop ebp
	ret

%define	gdtd	[ebp+8]

loadGDT:	;linux
;_loadGDT:	;mac
	push ebp
	mov ebp, esp
	pushad
	pushf
	
	mov eax, gdtd
	lgdt [eax]

	popf
	popad
	mov esp, ebp
	pop ebp
	ret

%define	idtd	[ebp+8]

loadIDT:	;linux
;_loadIDT:	;mac
	push ebp
	mov ebp, esp
	pushad
	pushf
	
	mov eax, idtd
	lidt [eax]

	popf
	popad
	mov esp, ebp
	pop ebp
	ret

%define gsv	[ebp+28]
%define fsv	[ebp+24]
%define esv	[ebp+20]
%define ssv	[ebp+16]
%define dsv	[ebp+12]
%define csv	[ebp+8]

setSelectors:	;linux
;_setSelectors:	;mac
	push ebp
	mov ebp, esp
	push eax
	pushf

	; set ds
	mov eax, dsv
	mov ds, eax

	mov eax, esv
	mov es, eax

	mov eax, fsv
	mov fs, eax

	mov eax, gsv
	mov gs, eax

	;set up ss
	mov eax, ssv
	mov ss, eax

	mov eax, csv
	push eax
	push _cont
	retf
_cont:
	popf
	pop eax
	mov esp, ebp
	pop ebp
	ret


go:
;_go:
	push dword [rnd]
	call dq
	pop ebx
	mov [cur], eax
	mov esp, [eax+4]
	mov ebx, [eax]
	mov ss, ebx
	pop gs
	pop fs
	pop es
	pop ds
	popad
	iret

schedule:
;_schedule:
  pushf
  push eax	
	mov al, 0x20
	out 0x20, al
	inc byte [tmr]
	cmp byte [tmr], 3
  pop eax
	je _switch
	popf
	iret
_switch:
	mov byte [tmr], 0
	popf
	pushad
	push ds
	push es
	push fs
	push gs
	mov eax, [cur]
	mov [eax+4], esp
	push eax
	push dword [rnd]
	call nq
	pop eax
	pop eax

	push dword [rnd]
	call dq
	pop ebx
	mov [cur], eax
	mov esp, [eax+4]
	mov ebx, [eax]
	mov ss, ebx
	pop gs
	pop fs
	pop es
	pop ds
	popad
	iret
_afterSwitch:
	ret

%define port [ebp+8]
%define write [ebp+12]

outportb:
;_outportb:
	push ebp
	mov ebp, esp
	push eax
	push edx
	pushf

	mov al, write
	mov dx, port
	out dx, al

	popf
	pop edx
	pop eax
	mov esp, ebp
	pop ebp
	ret

%define mutex [ebp+8]

mutex_lock:
	push ebp
	mov ebp, esp
	push eax
	push ebx
	pushf

;mutex will be in ebx, the current process will be in 
	cli
	mov ebx, mutex
	cmp [ebx], dword 0
	je _open
		;this saves the blocked process to the mutex's queue
		pushf
		push cs
		push _open
		pushad
		push ds
		push es
		push fs
		push gs
		mov eax, [cur]
		mov [eax+4], esp
		push eax
		add ebx, 4;
		push ebx ;addres of the mutex's queue
		call nq
		pop ebx
		pop eax

		;this starts the next process on the ready queue
		push dword [rnd]
		call dq
		pop ebx
		mov [cur], eax
		mov esp, [eax+4]
		mov ebx, [eax]
		mov ss, ebx
		pop gs
		pop fs
		pop es
		pop ds
		popad
		iret
_open:
	mov ebx, mutex
	mov [ebx], dword 1
	;sti
	
	popf
	pop ebx
	pop eax
	mov esp, ebp
	pop ebp
	ret

mutex_unlock:
	push ebp
	mov ebp, esp
	push eax
	push ebx
	push ecx
	pushf

	cli
	mov ebx, mutex ;ebx has addr of mutex
	mov eax, ebx 
	add eax, 4 ;eax has addr of mutex queue
	mov ecx, [eax]
	cmp dword [eax+4], ecx
	je _close
		;this starts the next process on the ready queue
		push eax ; address of the queue
		call dq
		pop ebx

		push eax
		push dword [rnd]
		call nq
		; if I pop stuff here, it all crashes
		;pop ebx
		;pop ebx
		jmp _afterzero

_close:
	mov ebx, mutex
	mov [ebx], dword 0
_afterzero:
	;sti

	popf
	pop ecx
	pop ebx
	pop eax
	mov esp, ebp
	pop ebp
	ret

%define sema [ebp+8]

sema_down:
	push ebp
	mov ebp, esp
	push eax
	push ebx
	pushf

;mutex will be in ebx, the current process will be in 
	cli
	mov ebx, sema
	sub [ebx], dword 1
	cmp [ebx], dword 0
	jge _down
		;this saves the blocked process to the sema's queue
		pushf
		push cs
		push _down
		pushad
		push ds
		push es
		push fs
		push gs
		mov eax, [cur]
		mov [eax+4], esp
		push eax
		add ebx, 4
		push dword ebx ;addres of the sema's queue
		call nq
		pop ebx
		pop eax

		;this starts the next process on the ready queue
		push dword [rnd]
		call dq
		pop ebx
		mov [cur], eax
		mov esp, [eax+4]
		mov ebx, [eax]
		mov ss, ebx
		pop gs
		pop fs
		pop es
		pop ds
		popad
		iret
_down:
	;sti
	
	popf
	pop ebx
	pop eax
	mov esp, ebp
	pop ebp
	ret


sema_up:
	push ebp
	mov ebp, esp
	push eax
	push ebx
	push ecx
	pushf

	cli
	mov ebx, mutex ;ebx has addr of mutex
	add [ebx], dword 1
	cmp [ebx], dword 0
	jg _up
		;this starts the next process on the ready queue
		add ebx, 4
		push dword ebx ; address of the queue
		call dq
		pop ebx

		push eax
		push dword [rnd]
		call nq
		pop eax
		pop eax

_up:
	;sti

	popf
	pop ecx
	pop ebx
	pop eax
	mov esp, ebp
	pop ebp
	ret
