#include "boot2.h"

gdt_entry_t gdt[5];
gdtr gdtd;

idt_entry_t idt[256];
gdtr idtd;

uint32_t stacks[10][1024];
int nextStack=0;

pcb pcbs[10];
int nextProcess=0;

rrQ *Round;
pcb *curr;

int timer;

mutex_t mute;

char *buffer;

char *buffet;

sema_t emptySlots;
sema_t fullSlots;

int nextWrite;
int nextRead;
int buffetPlace;

#define CODE_SEL 0x00000008
#define DATA_SEL 0x00000010
#define STACK_SEL 0x00000018
#define VIDEO_SEL 0x00000020
#define STACK_SIZE 1024


int main(){

	clearScr();
	writeScr("Initializing OS", 0, 0);
	writeScr("Setting up OS descriptors...", 1, 0);

	initGDTEntry(gdt, 0, 0, 0, 0);
	initGDTEntry(&gdt[1], 0, 655359, 0x9a, 0x40);
	initGDTEntry(&gdt[2], 0, 655359, 0x92, 0x40);
	initGDTEntry(&gdt[3], 0, 655359, 0x92, 0x40);
	initGDTEntry(&gdt[4], 0xb8000, (80*25*2)-1, 0x92, 0x40);

	gdtd.limit = (uint16_t)(sizeof(gdt) - 1);
	gdtd.base = (uint32_t)gdt;


	loadGDT((uint32_t)&gdtd);
	setSelectors(CODE_SEL, DATA_SEL, STACK_SEL, VIDEO_SEL, DATA_SEL, DATA_SEL);

	pmWriteScr("                         done.", 2, 0);
	int i = 3;
	while(i < 25){
		pmWriteScr("********************************************************************************", i++, 0);
	}

	pmClearScr();
	pmWriteScr("Running one producer and five consumers.",0,0);
	Round->front=0;
	Round->back=0;	//to show the queue is empty

	
	idtd.limit = (uint16_t)(sizeof(idt) -1);
	idtd.base = (uint32_t)idt;

	mutex_init(&mute);
	sema_init(&emptySlots,80);
	sema_init(&fullSlots,0);
	
	nextRead = 0;
	nextWrite = 0;
	buffetPlace = 0;
	
	buffer = "                                                                                ";
	buffet = "7284759204916184050138500188364651829050183756375498209754306";
	//buffet is 60 long

	uint32_t s;
	s = allocStack();

	initIDT();
	asm("cli");
	setupPIC();

	createProcess(DATA_SEL, STACK_SEL, (uint32_t) (s + STACK_SIZE), CODE_SEL, VIDEO_SEL, (uint32_t) producer1);
	s = allocStack();
	createProcess(DATA_SEL, STACK_SEL, (uint32_t) (s + STACK_SIZE), CODE_SEL, VIDEO_SEL, (uint32_t) consumer1);
	s = allocStack();
	createProcess(DATA_SEL, STACK_SEL, (uint32_t) (s + STACK_SIZE), CODE_SEL, VIDEO_SEL, (uint32_t) consumer2);
	s = allocStack();
	createProcess(DATA_SEL, STACK_SEL, (uint32_t) (s + STACK_SIZE), CODE_SEL, VIDEO_SEL, (uint32_t) consumer3);
	s = allocStack();
	createProcess(DATA_SEL, STACK_SEL, (uint32_t) (s + STACK_SIZE), CODE_SEL, VIDEO_SEL, (uint32_t) consumer4);
	s = allocStack();
	createProcess(DATA_SEL, STACK_SEL, (uint32_t) (s + STACK_SIZE), CODE_SEL, VIDEO_SEL, (uint32_t) consumer5);
/*	s = allocStack();
	createProcess(DATA_SEL, STACK_SEL, (uint32_t) (s + STACK_SIZE), CODE_SEL, VIDEO_SEL, (uint32_t) p7);
	s = allocStack();
	createProcess(DATA_SEL, STACK_SEL, (uint32_t) (s + STACK_SIZE), CODE_SEL, VIDEO_SEL, (uint32_t) p8);
	s = allocStack();
	createProcess(DATA_SEL, STACK_SEL, (uint32_t) (s + STACK_SIZE), CODE_SEL, VIDEO_SEL, (uint32_t) p9);
	s = allocStack();
	createProcess(DATA_SEL, STACK_SEL, (uint32_t) (s + STACK_SIZE), CODE_SEL, VIDEO_SEL, (uint32_t) p10);
*/
	go();
	//while(1)
	//return 0;
}

/// Round-robin functions ////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
void enqueue(rrQ *Q, pcb *proc){
	Q->queue[Q->back] = proc;
	Q->back = (Q->back + 1) % 11;
}

pcb *dequeue(rrQ *Q){
	pcb *ret = Q->queue[Q->front];
	Q->queue[Q->front] = 0;
	Q->front = (Q->front + 1) % 11;
	return ret;
}
///////////////////////////////////////////////////////////////////////////////////
/// Mutex/Sema functions //////////////////////////////////////////////////////////
void mutex_init(mutex_t *m){
	rrQ q;
	m->lock=0;
	q=m->Q;
	q.front=0;
	q.back=0;
}

void sema_init(sema_t *s, int count){
	s->available = count;
	s->Q.front = 0;
	s->Q.back = 0;
}
//////////////////////////////////////////////////////////////////////////////////	
//rest of functions
void clearScr(void){
	int i = 0;
	for(i = 0; i < 25; i++){
		writeScr("                                                                                ",i,0);
	}	
}

void pmClearScr(){
	int i = 0;
	for(i = 0; i < 25; i++){
		pmWriteScr("                                                                                ",i,0);
	}	
}

void writeScr(char *string, int row, int col){
	write_scr(string, row, col);	
}

void initGDTEntry(gdt_entry_t *entry, uint32_t baseAddress, uint32_t limit, uint8_t access, uint8_t attributes){
	entry->limit = (uint16_t) limit;
	entry->baseLow = (uint16_t) baseAddress;
	entry->baseMid = (uint8_t) (baseAddress >> 16);
	entry->access = access;
	entry->attr_lim = attributes | ((limit >> 16) & 0x0F);
	entry->baseHigh = (uint8_t) (baseAddress >> 24);
}

void initIDTEntry(idt_entry_t *entry, uint32_t base, uint16_t selector, uint8_t access){
	entry->zero = (uint8_t)0;
	entry->selector = selector;
	entry->access = access;
	entry->baseLow = (uint16_t)(base & 0X0000FFFF);
	entry->baseHigh = (uint16_t)(base >> 16);
}

void initIDT(){
	int i;
	for(i = 0; i < 32; i++)
		initIDTEntry(&idt[i],(uint32_t)stdHandle,CODE_SEL,0x8e);
	initIDTEntry(&idt[32],(uint32_t)schedule,CODE_SEL,0x8e);
	for(i = 33; i < 255; i++)
		initIDTEntry(&idt[i], 0, 0, 0 );
	loadIDT((uint32_t)&idtd);
}

uint32_t allocStack(){
	uint32_t ret = (uint32_t) &stacks[0][0];	// get the address of the start of the array
	uint32_t temp = STACK_SIZE;	// about to do some math
	temp *= nextStack;		// multiplies by the current stack setting to get offset
	ret += temp;
	nextStack++;			//get ready for next call
	return ret;
}

pcb *allocatePCB(){
	pcb *ret;
	ret = pcbs + (sizeof(pcb)*nextProcess);
	nextProcess++;
	return ret;
}

int createProcess(uint32_t ds, uint32_t ss, uint32_t stackTop, uint32_t cs, uint32_t es, uint32_t proccessEntry){
	uint32_t *st;
	st = (uint32_t *)stackTop;
	st--;
	*st = (uint32_t) 0x200;
	st--;
	*st = cs;
	st--;
	*st = proccessEntry;
	st--;
	*st = 0;   //ebp
	st--;
	*st = 0;   //esp
	st--;
	*st = 0;   //edi
	st--;
	*st = 0;   //esi
	st--;
	*st = 0;   //edx
	st--;
	*st = 0;   //ecx
	st--;
	*st = 0;   //ebx
	st--;
	*st = 0;   //eax
	st--;
	*st = ds;
	st--;
	*st = es;
	st--;
	*st = ds;  //fs
	st--;
	*st = ds;  //gs
	
	pcb *proc = allocatePCB();
	proc->ss = ss;
	proc->esp = (uint32_t) st;
	enqueue(Round, proc);
	return 1;
}

void stdHandle(){
	pmClearScr();
	pmWriteScr("Illegal operation performed.",0,0);
	while(1);
}

void setupPIC(){
	outportb(0x20, 0x11); // start 8259 master initialization
	outportb(0xA0, 0x11); // start 8259 slave initialization
	outportb(0x21, 0x20); // set master base interrupt vector (idt 32-38)
	outportb(0xA1, 0x28); // set slave base interrupt vector (idt 39-45)
	outportb(0x21, 0x04); // set cascade ...
	outportb(0xA1, 0x02); // on IRQ2
	outportb(0x21, 0x01); // finish 8259 initialization
	outportb(0xA1, 0x01);
	outportb(0x21, 0x0);
	outportb(0xA1, 0x0);
	outportb(0x21, 0xfe); // Turn on the clock IRQ
	outportb(0xa1, 0xff); // Turn off all others
}

////////////////////////////////
/// PROCESSES
///////////////////////////////
void producer1(){
	char *producer = "Producing:     "; 
	char produced;
	while(1){
	sema_down(&emptySlots);
	  produced = buffet[buffetPlace];
	  buffetPlace = (buffetPlace + 1) % 60;
	  *(producer + 11) = produced;
	  pmWriteScr(producer, 6, 0);
	mutex_lock(&mute);
	  buffer[nextWrite] = produced;
	  nextWrite = (nextWrite + 1) % 80;
	  pmWriteScr(buffer, 3, 0);
	mutex_unlock(&mute);
	sema_up(&fullSlots);
	}
}
void consumer1(){
	char *consumer = "Consuming:     ";
	char consumed;
	while(1){
	sema_down(&fullSlots);
	mutex_lock(&mute);
	  consumed = buffer[nextRead];
	  buffer[nextRead] = ' ';
	  nextRead = (nextRead + 1) % 80;
	  pmWriteScr(buffer, 3, 0);
	mutex_unlock(&mute);
	sema_up(&emptySlots);
	  *(consumer + 11) = consumed;
	  pmWriteScr(consumer, 8, 0);
	}
}
void consumer2(){
	char *consumer = "Consuming:     ";
	char consumed;
	while(1){
	sema_down(&fullSlots);
	mutex_lock(&mute);
	  consumed = buffer[nextRead];
	  buffer[nextRead] = ' ';
	  nextRead = (nextRead + 1) % 80;
	  pmWriteScr(buffer, 3, 0);
	mutex_unlock(&mute);
	sema_up(&emptySlots);
	  *(consumer + 11) = consumed;
	  pmWriteScr(consumer, 9, 0);
	}
}
void consumer3(){
	char *consumer = "Consuming:     ";
	char consumed;
	while(1){
	sema_down(&fullSlots);
	mutex_lock(&mute);
	  consumed = buffer[nextRead];
	  buffer[nextRead] = ' ';
	  nextRead = (nextRead + 1) % 80;
	  pmWriteScr(buffer, 3, 0);
	mutex_unlock(&mute);
	sema_up(&emptySlots);
	  *(consumer + 11) = consumed;
	  pmWriteScr(consumer, 10, 0);
	}
}
void consumer4(){
	char *consumer = "Consuming:     ";
	char consumed;
	while(1){
	sema_down(&fullSlots);
	mutex_lock(&mute);
	  consumed = buffer[nextRead];
	  buffer[nextRead] = ' ';
	  nextRead = (nextRead + 1) % 80;
	  pmWriteScr(buffer, 3, 0);
	mutex_unlock(&mute);
	sema_up(&emptySlots);
	  *(consumer + 11) = consumed;
	  pmWriteScr(consumer, 11, 0);
	}
}
void consumer5(){
	char *consumer = "Consuming:     ";
	char consumed;
	while(1){
	sema_down(&fullSlots);
	mutex_lock(&mute);
	  consumed = buffer[nextRead];
	  buffer[nextRead] = ' ';
	  nextRead = (nextRead + 1) % 80;
	  pmWriteScr(buffer, 3, 0);
	mutex_unlock(&mute);
	sema_up(&emptySlots);
	  *(consumer + 11) = consumed;
	  pmWriteScr(consumer, 12, 0);
	}
}
void p3(){
	int i = 0;
	char *msg3 = "Process 3 :              ";
	while(1){
		i = 0;
		for(i = 0; i < 10; i++){
			*(msg3 + 12 + i) = 'l';
			pmWriteScr(msg3,7,0);
		}
		for(i = 0; i < 10; i++)
			*(msg3 + 12 + i) = ' ';
	}
}
void p4(){
	int i = 0;
	char *msg4 = "Process 4 :              ";
	while(1){
		i = 0;
		for(i = 0; i < 10; i++){
			*(msg4 + 12 + i) = '~';
			pmWriteScr(msg4,8,0);
		}
		for(i = 0; i < 10; i++)
			*(msg4 + 12 + i) = ' ';
	}
}
void p5(){
	int i = 0;
	char *msg5 = "Process 5 :              ";
	while(1){
		i = 0;
		for(i = 0; i < 10; i++){
			*(msg5 + 12 + i) = 'm';
			pmWriteScr(msg5,9,0);
		}
		for(i = 0; i < 10; i++)
			*(msg5 + 12 + i) = ' ';
	}
}
void p6(){
	int i = 0;
	char *msg6 = "Process 6 :              ";
	while(1){
		i = 0;
		for(i = 0; i < 10; i++){
			*(msg6 + 12 + i) = '5';
			pmWriteScr(msg6,10,0);
		}
		for(i = 0; i < 10; i++)
			*(msg6 + 12 + i) = ' ';
	}
}
void p7(){
	int i = 0;
	char *msg7 = "Process 7 :              ";
	while(1){
		i = 0;
		for(i = 0; i < 10; i++){
			*(msg7 + 12 + i) = '|';
			pmWriteScr(msg7,11,0);
		}
		for(i = 0; i < 10; i++)
			*(msg7 + 12 + i) = ' ';
	}
}
void p8(){
	int i = 0;
	char *msg8 = "Process 8 :              ";
	while(1){
		i = 0;
		for(i = 0; i < 10; i++){
			*(msg8 + 12 + i) = 'w';
			pmWriteScr(msg8,12,0);
		}
		for(i = 0; i < 10; i++)
			*(msg8 + 12 + i) = ' ';
	}
}
void p9(){
	int i = 0;
	char *msg9 = "Process 9 :              ";
	while(1){
		i = 0;
		for(i = 0; i < 10; i++){
			*(msg9 + 12 + i) = '^';
			pmWriteScr(msg9,13,0);
		}
		for(i = 0; i < 10; i++)
			*(msg9 + 12 + i) = ' ';
	}
}
void p10(){
	int i = 0;
	char *msg10 = "Process 10:              ";
	while(1){
		i = 0;
		for(i = 0; i < 10; i++){
			*(msg10 + 12 + i) = 'P';
			pmWriteScr(msg10,14,0);
		}
		for(i = 0; i < 10; i++)
			*(msg10 + 12 + i) = ' ';
	}
}
