typedef unsigned char uint8_t;
typedef unsigned short uint16_t; 
typedef unsigned int uint32_t; 

struct gdtr_s{
	uint16_t limit;
	uint32_t base;
}__attribute__((packed));
typedef struct gdtr_s gdtr;

struct gdt_entry{
	uint16_t limit;
	uint16_t baseLow;
	uint8_t baseMid;
	uint8_t access;
	uint8_t attr_lim;
	uint8_t baseHigh;
}__attribute__((packed));
typedef struct gdt_entry gdt_entry_t;

struct idt_entry{
	uint16_t baseLow;
	uint16_t selector;
	uint8_t	zero;
	uint8_t access;
	uint16_t baseHigh;
}__attribute__((packed));
typedef struct idt_entry idt_entry_t;

struct pcb_s{
	uint32_t ss;
	uint32_t esp;
};
typedef struct pcb_s pcb;

struct rrQ_s{
	int front;
	int back;
	pcb *queue[11];
}__attribute__((packed));
typedef struct rrQ_s rrQ;

struct mutex_s{
	uint32_t lock;
	rrQ Q;
}__attribute__((packed));
typedef struct mutex_s mutex_t;

struct sema_s{
	uint32_t available;
	rrQ Q;
}__attribute__((packed));
typedef struct sema_s sema_t;

//round robbin functions
void enqueue(rrQ *Q, pcb *proc);
pcb *dequeue(rrQ *Q);

//mutex functions
void mutex_init(mutex_t *m);
void mutex_lock(mutex_t *m);
void mutex_unlock(mutex_t *m);

//semaphore functions
void sema_init(sema_t *s, int count);
void sema_down(sema_t *s);
void sema_up(sema_t *s);

//C functions
void clearScr(void);
void pmClearScr(void);
void writeScr(char *string, int row, int col);
void initGDTEntry(gdt_entry_t *entry, uint32_t baseAddress, uint32_t limit, uint8_t access, uint8_t attributes);
void initIDTEntry(idt_entry_t *entry, uint32_t base, uint16_t selector, uint8_t access);
void initIDT();
uint32_t allocStack();
pcb *allocatePCB();
int createProcess(uint32_t ds, uint32_t ss, uint32_t stackTop, uint32_t cs, uint32_t es, uint32_t proccessEntry);
void stdHandle();
void setupPIC();

//Assembly functions
void write_scr(char *string, int row, int col);
void pmWriteScr(char *string, int row, int col);
void loadGDT(uint32_t addr);
void loadIDT(uint32_t addr);
void setSelectors(uint32_t cs, uint32_t ds, uint32_t ss, uint32_t es, uint32_t fs, uint32_t gs);
void go();
void schedule();
void outportb(uint16_t port, uint8_t byte);


//Processes
void producer1();
void consumer1();
void consumer2();
void consumer3();
void consumer4();
void consumer5();
void p3();
void p4();
void p5();
void p6();
void p7();
void p8();
void p9();
void p10();

